/*
 * Read ARM PMU Register
 * Author: Yiquan Chen
 * http://www.ilinuxkernel.com
 * ilinuxkernel@gmail.com 
 */
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>

/* All counters, including PMCCNTR_EL0, are disabled/enabled */

#define QUADD_ARMV8_PMCR_E      (1 << 0)
/* Reset all event counters, not including PMCCNTR_EL0, to 0 */
#define QUADD_ARMV8_PMCR_P      (1 << 1)
/* Reset PMCCNTR_EL0 to 0 */
#define QUADD_ARMV8_PMCR_C      (1 << 2)
/* Clock divider: PMCCNTR_EL0 counts every clock cycle/every 64 clock cycles */
#define QUADD_ARMV8_PMCR_D      (1 << 3)
/* Export of events is disabled/enabled */
#define QUADD_ARMV8_PMCR_X      (1 << 4)
/* Disable cycle counter, PMCCNTR_EL0 when event counting is prohibited */
#define QUADD_ARMV8_PMCR_DP     (1 << 5)
/* Long cycle count enable */
#define QUADD_ARMV8_PMCR_LC     (1 << 6)

#define	ARMV8_PMCR_MASK		0x3f	 /* Mask for writable bits */

static inline unsigned int armv8_pmu_pmcr_read(void)
{
        unsigned int val;
        /* Read Performance Monitors Control Register */
        asm volatile("mrs %0, pmcr_el0" : "=r" (val));
        return val;
}
static inline void armv8_pmu_pmcr_write(unsigned int val)
{
    asm volatile("msr pmcr_el0, %0" : :"r" (val & ARMV8_PMCR_MASK));
}


static inline  long long armv8_read_CNTPCT_EL0(void)
{
   long long val;
   asm volatile("mrs %0, CNTVCT_EL0" : "=r" (val));

   return val;
}

static void enable_all_counters(void)
{
 
    return;
    unsigned int val;
    /* Enable all counters */
    val = armv8_pmu_pmcr_read();
    val |= QUADD_ARMV8_PMCR_E | QUADD_ARMV8_PMCR_X;
    armv8_pmu_pmcr_write(val);
}

static void reset_all_counters(void)
{

   return ; 
   unsigned int val;
    val = armv8_pmu_pmcr_read();
    val |= QUADD_ARMV8_PMCR_P | QUADD_ARMV8_PMCR_C;
    armv8_pmu_pmcr_write(val);
}

static unsigned int enabled=0;

unsigned int readticks(unsigned int *result)
{
    struct timeval t;
    unsigned int cc;
    unsigned int val;
    if (!enabled) {
        reset_all_counters();
        enable_all_counters();
        enabled = 1;
    }
    cc = armv8_pmu_pmcr_read();
    gettimeofday(&t,(struct timezone *) 0);
    result[0] = cc;
    result[1] = t.tv_usec;
    result[2] = t.tv_sec;
  
    return cc;
}


static inline unsigned int armv8pmu_pmcr_read(void)
{
	unsigned int val;
	asm volatile("mrs %0, pmcr_el0" : "=r" (val));
	return val;
}

#define u32 unsigned int
#define u64 unsigned long long
#define isb()		asm volatile("isb" : : : "memory")

static inline u64 arch_counter_get_cntpct(void)
{
	u64 cval;

	isb();
        asm volatile("mrs %0, PMCCNTR_EL0" : "+r"(cval));
	return cval;
}

int main()
{

  unsigned int start,end;
  unsigned int result[3]; 
  unsigned long long timer;
  u32 pmcr_el;

  pmcr_el = armv8pmu_pmcr_read();
  printf("\nPMCR_EL0 Register:%lX ", pmcr_el);
  timer = arch_counter_get_cntpct();
  printf("\nCPU Cycle Count:0x%llX ",timer);  
  sleep(5);
  timer = arch_counter_get_cntpct();
  printf("\nCPU Cycle Count:0x%llX \n",timer);
  asm volatile("mrs %0, PMOVSCLR_EL0" : "=r"(pmcr_el));
  printf(" Register PMOVSCLR_EL0:0x%lX \n", pmcr_el); 
  
  asm volatile("mrs %0, pmuserenr_el0" : "=r"(pmcr_el));
  printf(" Register pmuserenr_el0:0x%lX \n", pmcr_el);

  asm volatile("mrs %0, PMCNTENSET_EL0" : "=r"(pmcr_el));
  printf(" Register PMCNTENSET_EL0:0x%lX \n", pmcr_el);

  asm volatile("mrs %0, PMCCFILTR_EL0" : "=r"(pmcr_el));
  printf(" Register PMCCFILTR_EL0:0x%lX \n", pmcr_el);

  asm volatile("mrs %0, PMCNTENCLR_EL0" : "=r"(pmcr_el));
  printf(" Register PMCNTENCLR_EL0:0x%lX \n", pmcr_el);

  asm volatile("mrs %0, PMOVSSET_EL0" : "=r"(pmcr_el));
  printf(" Register PMOVSSET_EL0:0x%lX \n", pmcr_el);

  return 0;

}
